/*******************************************************************************
rccrypt Version 1.1
********************************************************************************
This program implements rc5 128 bit block cipher.

Copyright (C) 2001 R.G.Jones

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email:rich@ricksoft.co.uk
********************************************************************************
********************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <pwd.h>
#include <time.h>

/********************************************************************************/
/*Define the magic numbers used for RC5-128*/
#define P64 0xb7e151628aed2a6b /* P64=(e-2)*2^64 */
#define Q64 0x9e3779b97f4a7c15 /* Q64=(L-2)*2^64 where L is Golden Ratio*/
/*Other Definitions*/
#define ULLONG unsigned long long
#define ULONG unsigned long
enum boolean {FALSE,TRUE};

/********************************************************************************/
/* The following are definitions of subroutines defined after main*/
void bail (char *c);
ULLONG remainder(ULLONG numerator, ULLONG divisor);
long max(long a,long b);
ULLONG lrotate(ULLONG num,ULLONG n);
ULLONG rrotate(ULLONG num,ULLONG n);
char get_random_char();

/********************************************************************************/
/* Main program begins here*/
/********************************************************************************/
int main(int argc, char **argv)
{
/********************************************************************************/
/*Declare all variables here*/
struct passwd *user_data;	/*will contain details of user data*/
int d;	/*used for determining command line options*/
int one_char;	/*used for getting chars from the input stream*/
long i;	/*used for loop control*/
int decode;
FILE *input;	/*either from a file or the standard input*/
FILE *output;	/*either to a file or the standard ouput*/
FILE *keyfile;	/*used to input the key if it is in a file*/
char key_buffer[513];	/*to hold key if input from a file*/
char my_cwd[500];	/*to hold the directory the program is started in*/
unsigned char *padded_keystring;	/*Pointer to padded array for the key to fit into*/
unsigned char sub_array[9];		/*array to get chunks out of the padded array*/
ULLONG *K;	/*Pointer for integer access to key*/
long c;		/*number of 64 bit integers in the key*/
long rounds;	/*number of rounds required (0..255)*/
long t;		/*number of subkeys:2*(rounds+1)*/
ULLONG *S;	/*Pointer to subkey array*/
long j,k;	/*used to access arrays*/
long num_chars_in_keystring;
long num_chars_in_padded_keystring;
long num_lls_in_key;
char *k_chars;	/*points to key string given in command line*/
char *f_chars;  /*Points to file containing key*/
char *r_chars;	/*points to rounds string given in cammand line*/
char *infile;	/*points to filename if given on command line*/
char *outfile;	/*points to filename if given on command line*/
char *tail;	/*for char to number conversion checking*/
ULONG temp;	/*used for VERY temporary storage of a number...*/
ULLONG A,B;	/*64-bit integers used to mangle sub keys*/
char buffer[17];	/*buffer to read in data one longer than the data so we can print it*/
size_t count;		/*how many items were read from the stream*/
ULLONG *L,*R;		/*pointers to parts of the buffer-left and right ULLONG chunks*/

/********************************************************************************/
/*Create a random seed based on the calendar time*/
srandom(time(NULL));

/********************************************************************************/
/*Set all variables etc to default/null values*/
opterr = 0;num_chars_in_keystring=0;rounds=10;
decode=FALSE;
k_chars=NULL;
f_chars=NULL;
r_chars=NULL;
keyfile=NULL;
infile=NULL;
outfile=NULL;
/********************************************************************************/
/*Check out what options have been given and set variables*/
while ((d = getopt (argc, argv, "dk:f:r:i:o:")) != -1)
    switch (d)
      {
      case 'o':
      	outfile=optarg;
	break;
      case 'i':
      	infile=optarg;
	break;
      case 'd':
      	decode=TRUE;
	break;
      case 'k':
        k_chars = optarg;
        break;
      case 'f':
        f_chars = optarg;
        break;
      case 'r':
        r_chars = optarg;
	/*set the number of rounds*/
	errno=0;
	rounds=strtol(r_chars,&tail,10);
	if ((tail==r_chars) || (errno) || (rounds>255) || (rounds<0))bail("ERROR: Rounds must be 0..255");
        break;
      case '?':
        fprintf(stderr,"Usage:\n\t%s\t [-d] [-r rounds ] [ -k key | -f keyfile] [ -i infilename ] [ -o outfilename ]\n\n",argv[0]);
	fprintf(stderr,"\nSee README file for further information\n");
        return 1;
      default:
      }
/********************************************************************************/
/* If a key has not been given, use the file given or ~/.rccrypt_key*/
if(k_chars == NULL)
        {
        if (NULL==f_chars)
                {
                /*remember what directory we are in*/
                if(getcwd(my_cwd,500) != NULL)
                        {
                        /*find the user's home directory*/
                        user_data=getpwuid(getuid());
                        if(user_data == NULL) bail("Unable to determine user's home directory");
                        if(user_data->pw_dir != NULL)
                                {
                                if (chdir(user_data->pw_dir)) bail("Unable to change directory");
                                keyfile=fopen(".rccrypt_key","r");
                                /*return to the original directory*/
                                if (chdir(my_cwd)) bail("Unable to change directory");
                                }
                        }
                else
                        {
                        bail("Unable to determine current directory.");
                        }
                }
        else
                {
                keyfile=fopen(f_chars,"r");
                }
        /*read in the key if the file exists*/
        if (keyfile != NULL)
                {
                count=fread (key_buffer,1,512,keyfile);
                /*make the end null*/
                key_buffer[count-1]='\0';
                /*set the k_chars to point to the buffer*/
                k_chars=key_buffer;
                /*close the file*/
                fclose(keyfile);
                }
        /*note - if get to here without creating key, 0x00000000will be used*/
        }
/********************************************************************************/
/*set the number of chars in the key*/
if(k_chars != NULL)
	{
	num_chars_in_keystring=strlen(k_chars);
	}

/********************************************************************************
Converting the key....
Each 2 chars in the key is one byte.
One long has 8 chars.
One Long long has 16 chars.
Create a string containing the chars in the key, padded with enough zeros
at the beginning to make it up to an integer number of long longs.
Create the final long long array the right size.
The use atoul(base 16) to convert the char array 8 chars at a time into the final array.
********************************************************************************/
num_lls_in_key=num_chars_in_keystring/16;
if (remainder(num_chars_in_keystring,16)>0) num_lls_in_key +=1;
c=num_lls_in_key;
num_chars_in_padded_keystring=num_lls_in_key*16;
/*create the padded array*/
padded_keystring=(char *)malloc(num_chars_in_padded_keystring+1);
if (padded_keystring==NULL) bail("Malloc error:padded_keystring");
/*fill the array with zeros then copy the key to the end of it*/
memset(padded_keystring,'0',num_chars_in_padded_keystring);
padded_keystring[num_chars_in_padded_keystring]='\0';/*make sure it finishes with a null char so can use string fns on it*/
strncpy(padded_keystring+num_chars_in_padded_keystring-num_chars_in_keystring,k_chars,num_chars_in_keystring);
/*create the long long array for the key*/
K=(ULLONG *)malloc(num_lls_in_key*8);
if (K==NULL) bail("Malloc error:K");
/*fill it with zeros*/
memset(K,0,num_lls_in_key*8);
/*now convert each chunk of 8 chars to a long and copy it into the array*/
for (i=0,j=0;i<num_chars_in_padded_keystring;i+=8,j+=4)
	{
	strncpy(sub_array,padded_keystring+i,8);sub_array[8]='\0';
	errno=0;
	temp=strtoul(sub_array,&tail,16);
	if (tail!=(char *)(sub_array+8)) bail("Key must be a hex number");
	if(errno) bail("Key: overflow on conversion");
	memcpy(((char *)K)+j,(char *)&temp,4);/*copy the ULONG into the string*/
	}
/********************************************************************************/
/*Now fill the subkeys with pseudo random bits*/
t=2*(rounds+1);
S=(ULLONG *)malloc(t*8);
if (S==NULL) bail("Malloc Error:S");
S[0]=P64;
for(i=1;i<t;i++)
	{
	S[i]=S[i-1]+Q64;
	}
/*Mangle the subkeys with the Key*/
k=j=0;
A=B=0;
for(i=0;i<3*max(t,c);i++)
	{
	S[j]=lrotate(S[j]+A+B,3);
	K[k]=lrotate(K[k]+A+B,A+B);
	A=S[j];
	B=K[k];
	if (++j>=t) j=0;
	if(++k>=c) k=0;
	}
/********************************************************************************/
/*Determine where our input will be from*/
if (infile == NULL)
	{
	input=stdin;/*get info from the standard input*/
	/*this allows a user to pipe stuff in...*/
	}
else
	{
	input=fopen(infile,"r");
	if(input==NULL) bail("Failed to open input file");
	}
/********************************************************************************/
/*Determine where the output goes to*/
if (outfile==NULL)
	{
	output=stdout; /*send output to the standard output*/
	/*allows a user to pipe or redirect stuff out*/
	}
else
	{
	output=fopen(outfile,"w");
	if(output==NULL) bail("Failed to open output file");
	}
/********************************************************************************
Now we can do some encryption
We need to get 2 64-bit words for each block to encrypt
if there is not enough data, we need to pad the words with extra bytes
this program uses random printable chars for padding
********************************************************************************/

one_char=getc(input);
while(one_char!=EOF)
	{
	ungetc(one_char,input);/*put back the character as it isn't an eof*/
	/*Read 16 bytes of data into the buffer*/
	/*If we don't get 16 bytes, pad the buffer with . to overwrite previous values*/
	count=fread(buffer,1,16,input);
	if (count<16)
		{
		for (i=count;i<16;i++) buffer[i]=get_random_char();
		}
	buffer[16]='\0';/*only needed if we print it directly*/
	/*now we have sizteen bytes ready to work on*/
	/*split it into two ULLONGs, left and right*/
	L=(ULLONG *)buffer;
	R=(ULLONG *)(buffer+8);
	if (decode)
		{
		/*decrypt the data...*/
		for (i=rounds;i>=1;i--)
			{
			*R=rrotate(*R-S[2*i+1],*L) ^ *L;
			*L=rrotate(*L-S[2*i],*R) ^ *R;
			}
		*L -= S[0];
		*R -= S[1];
		}
	else
		{
		/*run the data through the encryption the right number of times*/
		*L += S[0];
		*R += S[1];
		for (i=1;i<=rounds;i++)
			{
			*L=S[2*i]+lrotate(*L ^ *R,*R);
			*R=S[2*i+1]+lrotate(*R ^ *L,*L);
			}
		}
	/*now output the data we worked on*/
	count=fwrite(buffer,1,16,output);
	if (count != 16) bail("Error in writing to output stream\n");
	/*Get the next char so we can test it for eof*/
	one_char=getc(input);
	}	
/********************************************************************************/
/*Close file handlers to ensure all data is flushed*/
if (input != stdin) fclose(input);
if (output != stdout) fclose(output);
/********************************************************************************/
/*Free up resources*/
/*Not really needed as this is the end of the program, but it's good practise*/
free(padded_keystring);
free(K);
free(S);
return (0);
}/*end of Main*/
/********************************************************************************/
/********************************************************************************/
/********************************************************************************/
/*give the remainder of an integer division*/
ULLONG remainder(ULLONG numerator, ULLONG divisor)
{
if (divisor == 0)
	{
	bail("Divide by zero error in 'remainder'");
	}
while (numerator>=divisor)
	{
	numerator -= divisor;
	}
return numerator;
}

/********************************************************************************/
/*Exit the program on abnormal errors*/
void bail (char *c)
{
perror(c);
exit(1);
}

/********************************************************************************/
/*Return the maximum value of the two*/
long max(long a,long b)
{
if (a>b)
	{
	return a;
	}
else
	{
	return b;
	}
}

/********************************************************************************/
/*Rotate the ULLONG left by n places*/
ULLONG lrotate(ULLONG num,ULLONG n)
{
/*first, make sure n is mod 64 so we don't lose any bits...*/
n &= 0x3f;
/*right shift by 64-n, left shift by n then OR them together*/
return (num<<n) | (num>>(64-n));
}
/********************************************************************************/
/*Rotate the ULLONG right by n places*/
ULLONG rrotate(ULLONG num,ULLONG n)
{
/*first, make sure n is mod 64 so we don't lose any bits...*/
n &= 0x3f;
/*left shift by 64-n, right shift by n then OR them together*/
return (num>>n) | (num<<(64-n));
}
/********************************************************************************/
/*Return a random printable character*/
char get_random_char()
{
char c;
do
	{
	c=(char)(random() & 0xff);
	}while (!isprint((int)c));
return c;
}
